/*****************************************************************************
 * Copyright (c) 2020, rrcSmall FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "Instruction.h"

#include "Write.h"
#include "Copy.h"
#include "Load.h"
#include "Init.h"
#include "Calc.h"
#include "CalcImm.h"
#include "Branch.h"
#include "Poll.h"
#include "Loop.h"
#include "Jump.h"
#include "Return.h"
#include "Set.h"
#include "Wait.h"
#include "End.h"

namespace Instruction {

    std::unique_ptr<Instruction>
    Instruction::Instruction::decodeInstructionFromBytes(uint32_t offset, const std::vector<uint8_t> &bytes) {
        auto result = std::unique_ptr<Instruction>(nullptr);

        CommandOp command = getCommandFromByte(bytes[offset]);
        switch (command) {
            case CommandOp::WRITE:
                result = std::make_unique<Write>();
                break;
            case CommandOp::COPY:
                result = std::make_unique<Copy>();
                break;
            case CommandOp::LOAD:
                result = std::make_unique<Load>();
                break;
            case CommandOp::INIT:
                result = std::make_unique<Init>();
                break;
            case CommandOp::CALC:
                result = std::make_unique<Calc>();
                break;
            case CommandOp::CALC_IMM:
                result = std::make_unique<CalcImm>();
                break;
            case CommandOp::BRANCH:
                result = std::make_unique<Branch>();
                break;
            case CommandOp::POLL:
                result = std::make_unique<Poll>();
                break;
            case CommandOp::LOOP:
                result = std::make_unique<Loop>();
                break;
            case CommandOp::JUMP:
                result = std::make_unique<Jump>();
                break;
            case CommandOp::RETURN:
                result = std::make_unique<Return>();
                break;
            case CommandOp::SET:
                result = std::make_unique<Set>();
                break;
            case CommandOp::WAIT:
                result = std::make_unique<Wait>();
                break;
            case CommandOp::END:
                result = std::make_unique<End>();
                break;
        }

        if (result != nullptr) {
            result->fromBytes(offset, bytes);
        }

        return result;
    }

    Instruction::CommandOp Instruction::getCommandFromByte(uint8_t command) {
        if (((command >> 6) & 0b11) == 0b00) {
            return CommandOp::WRITE;
        } else if (((command >> 6) & 0b11) == 0b01) {
            return CommandOp::COPY;
        } else if (((command >> 3) & 0b11111) == 0b11000) {
            return CommandOp::LOAD;
        } else if (((command >> 3) & 0b11111) == 0b11001) {
            return CommandOp::INIT;
        } else if (((command >> 5) & 0b111) == 0b100) {
            return CommandOp::CALC;
        } else if (((command >> 5) & 0b111) == 0b101) {
            return CommandOp::CALC_IMM;
        } else if (((command >> 3) & 0b11111) == 0b11010) {
            return CommandOp::BRANCH;
        } else if (((command >> 3) & 0b11111) == 0b11011) {
            return CommandOp::POLL;
        } else if (((command >> 1) & 0b1111111) == 0b1110000) {
            return CommandOp::LOOP;
        } else if (command == 0b11101000) {
            return CommandOp::JUMP;
        } else if (((command >> 2) & 0b111111) == 0b111100) {
            return CommandOp::RETURN;
        } else if (((command >> 2) & 0b111111) == 0b111110) {
            return CommandOp::SET;
        } else if (command == 0xFE) {
            return CommandOp::WAIT;
        } else if (command == 0xFF) {
            return CommandOp::END;
        }

        return CommandOp::END;
    }
}

