/*****************************************************************************
 * Copyright (c) 2020, rrcSmall FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "Load.h"
#include "../Registers.h"
#include <iostream>
#include <iomanip>
#include "../OutputContext.h"

void Instruction::Load::fromBytes(uint32_t offset, const std::vector<uint8_t> &bytes) {
    _address = offset;

    uint32_t command = bytes[offset++];

    increment = (command >> 2) & 0b1;
    addressOffset = command & 0b11;

    address = bytes[offset++] << 16;
    address |= bytes[offset++] << 8;
    address |= bytes[offset++];

    offset++;


    uint32_t count = bytes[offset++] << 16;
    count |= bytes[offset++] << 8;
    count |= bytes[offset++];

    for (uint32_t i = 0; i < count; ++i) {
        uint32_t entry = bytes[offset++] << 24;
        entry |= bytes[offset++] << 16;
        entry |= bytes[offset++] << 8;
        entry |= bytes[offset++];
        data.emplace_back(entry);
    }


    _endAddress = offset;
}

std::string Instruction::Load::toString() const {
    std::stringstream op;

    if (addressOffset == 0 && increment == 0 &&
        address == (uint32_t) KnownRegisters::PCIE_MASTER_SPICO_FIRMWARE_SPECIAL_REGISTER) {
        op << "LOAD PCIE_MASTER_SPICO_FIRMWARE DATA 0x";
        for (auto d : data) {
            op << std::hex << std::setw(8) << std::setfill('0') << d;
        }
        op << "\n";
        auto fw = SmallFirmwareFormat::fromBSMLoad(data, 0);
        for (uint32_t i = 0; i < fw.firmware.size(); ++i) {
            op << "    SPICO_FW[" << std::hex << std::setw(4) << std::setfill('0') << i << "]";
            op << " = 0x" << std::hex << std::setw(3) << std::setfill('0') << fw.firmware[i] << "\n";
        }
    } else if (addressOffset == 0 && increment == 0 &&
               address == (uint32_t) KnownRegisters::PCIE_BROADCAST_SERDES_FIRMWARE_SPECIAL_REGISTER) {
        op << "LOAD PCIE_BROADCAST_SERDES_FIRMWARE DATA 0x";
        for (auto d : data) {
            op << std::hex << std::setw(8) << std::setfill('0') << d;
        }
        op << "\n";
        auto fw = SmallFirmwareFormat::fromBSMLoad(data, 0);
        for (uint32_t i = 0; i < fw.firmware.size(); ++i) {
            op << "    SERDES_FW[" << std::hex << std::setw(4) << std::setfill('0') << i << "]";
            op << " = 0x" << std::hex << std::setw(3) << std::setfill('0') << fw.firmware[i] << "\n";
        }
    } else {
        op << "LOAD INTO " << getAddressRegisterName(address, addressOffset);
        if (data.size() > 1 && increment) {
            op << " UNTIL " << getAddressRegisterName(address + data.size() - 1, addressOffset);
        }
        op << " INCREMENT " << std::dec << (uint32_t) increment;
        op << " DATA 0x";
        for (auto d : data) {
            op << std::hex << std::setw(8) << std::setfill('0') << d;
        }

        if (data.size() > 1) {
            for (uint32_t i = 0; i < data.size(); ++i) {
                op << "\n    " << getAddressRegisterName(address + (increment ? i : 0)) << " = 0x" << std::hex
                   << std::setw(8) << std::setfill('0') << data[i];
            }
        }
    }


    return op.str();
}

std::vector<Instruction::OutputFormat> Instruction::Load::toOutputFormat(const OutputContext &context) const {
    if (addressOffset == 0 && increment == 0 &&
        address == (uint32_t) KnownRegisters::PCIE_MASTER_SPICO_FIRMWARE_SPECIAL_REGISTER) {
        uint32_t offset = 0;

        std::vector<OutputFormat> f{
                OutputFormat{getAddress() + offset++ * 4,
                             dynamic_cast<std::stringstream &>(std::stringstream("") << "LOAD"
                                                                                     << " PCIE_MASTER_SPICO_FIRMWARE").str()}
        };

        f.emplace_back(OutputFormat{getAddress() + offset++ * 4,
                                    dynamic_cast<std::stringstream &>(std::stringstream("") << "    COUNT " << std::dec
                                                                                            << data.size()).str()});

        for (auto d : data) {
            f.emplace_back(OutputFormat{getAddress() + offset * 4,
                                        dynamic_cast<std::stringstream &>(std::stringstream("") << "    0x" << std::hex
                                                                                                << std::setw(8)
                                                                                                << std::setfill('0')
                                                                                                << d).str()});
            offset++;
        }

        return f;
    } else if (addressOffset == 0 && increment == 0 &&
               address == (uint32_t) KnownRegisters::PCIE_BROADCAST_SERDES_FIRMWARE_SPECIAL_REGISTER) {
        uint32_t offset = 0;

        std::vector<OutputFormat> f{
                OutputFormat{getAddress() + offset++ * 4,
                             dynamic_cast<std::stringstream &>(std::stringstream("") << "LOAD"
                                                                                     << " PCIE_BROADCAST_SERDES_FIRMWARE").str()},
        };

        f.emplace_back(OutputFormat{getAddress() + offset++ * 4,
                                    dynamic_cast<std::stringstream &>(std::stringstream("") << "    COUNT " << std::dec
                                                                                            << data.size()).str()});

        for (auto d : data) {
            f.emplace_back(OutputFormat{getAddress() + offset * 4,
                                        dynamic_cast<std::stringstream &>(std::stringstream("") << "    0x" << std::hex
                                                                                                << std::setw(8)
                                                                                                << std::setfill('0')
                                                                                                << d).str()});
            offset++;
        }

        return f;
    } else {
        uint32_t offset = 0;

        std::vector<OutputFormat> f{
                OutputFormat{getAddress() + offset++ * 4,
                             dynamic_cast<std::stringstream &>(std::stringstream("") << "LOAD" << " "
                                                                                     << context.getAddressRegisterName(
                                                                                             address, addressOffset)
                                                                                     << ", INC " << std::dec
                                                                                     << (uint32_t) increment).str()}
        };

        f.emplace_back(OutputFormat{getAddress() + offset++ * 4,
                                    dynamic_cast<std::stringstream &>(std::stringstream("") << "    COUNT " << std::dec
                                                                                            << data.size()).str()});

        for (auto d : data) {
            f.emplace_back(OutputFormat{getAddress() + offset * 4,
                                        dynamic_cast<std::stringstream &>(std::stringstream("") << "    "
                                                                                                << context.getAddressRegisterName(
                                                                                                        address +
                                                                                                        (increment ? offset - 2 : 0),
                                                                                                        addressOffset)
                                                                                                << " = 0x" << std::hex
                                                                                                << std::setw(8)
                                                                                                << std::setfill('0')
                                                                                                << d).str()});
            offset++;
        }


        return f;
    }
}


std::vector<uint32_t> Instruction::Load::getPossibleBranches() const {
    return std::vector<uint32_t>{_endAddress};
}

std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>>
Instruction::Load::execute(AnalysisState &state) const {

    if (addressOffset == 0 && increment == 0 &&
        address == (uint32_t) KnownRegisters::PCIE_MASTER_SPICO_FIRMWARE_SPECIAL_REGISTER) {
        state.pciSPICOFirmware = SmallFirmwareFormat::fromBSMLoad(data, 0);
    } else if (addressOffset == 0 && increment == 0 &&
               address == (uint32_t) KnownRegisters::PCIE_BROADCAST_SERDES_FIRMWARE_SPECIAL_REGISTER) {
        state.pciSerDesFirmware = SmallFirmwareFormat::fromBSMLoad(data, 0);
    } else {
        for (uint32_t i = 0; i < data.size(); ++i) {
            state.setRegister(address + (increment ? i : 0), data[i]);
        }
    }

    state.current = _endAddress;

    state.addKnownJump(getEndAddress(), getAddress(), JumpKind::Continue);

    return std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>>();
}

std::vector<uint8_t> Instruction::Load::toBytes() const {
    std::vector<uint8_t> bytes;

    bytes.emplace_back((uint8_t) getCommand() | (increment << 2) | addressOffset);

    bytes.emplace_back((address >> 16) & 0xFF);
    bytes.emplace_back((address >> 8) & 0xFF);
    bytes.emplace_back(address & 0xFF);

    bytes.emplace_back(0x00);

    bytes.emplace_back((data.size() >> 16) & 0xFF);
    bytes.emplace_back((data.size() >> 8) & 0xFF);
    bytes.emplace_back(data.size() & 0xFF);

    for (auto d : data) {
        bytes.emplace_back((d >> 24) & 0xFF);
        bytes.emplace_back((d >> 16) & 0xFF);
        bytes.emplace_back((d >> 8) & 0xFF);
        bytes.emplace_back(d & 0xFF);
    }

    return bytes;
}
