/*****************************************************************************
 * Copyright (c) 2020, rrcSmall FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "Loop.h"
#include "../Registers.h"
#include <iostream>
#include <iomanip>
#include "../OutputContext.h"

void Instruction::Loop::fromBytes(uint32_t offset, const std::vector<uint8_t> &bytes) {
    _address = offset;

    uint32_t command = bytes[offset++];

    counter = command & 0b1;

    jumpAddress = bytes[offset++] << 16;
    jumpAddress |= bytes[offset++] << 8;
    jumpAddress |= bytes[offset++];


    _endAddress = offset;
}

std::string Instruction::Loop::toString() const {
    std::stringstream op;
    op << "LOOP IF " << getRegisterName(static_cast<KnownRegisters>((uint32_t) KnownRegisters::BSM_COUNTER_0 + counter))
       << " != 0 JUMP 0x" << std::hex << std::setw(6) << std::setfill('0') << jumpAddress;
    return op.str();
}

std::vector<Instruction::OutputFormat> Instruction::Loop::toOutputFormat(const OutputContext &context) const {
    return std::vector<OutputFormat>{
            OutputFormat{getAddress(), dynamic_cast<std::stringstream &>(std::stringstream("") << "LOOP" << " "
                                                                                               << context.getAddressRegisterName(
                                                                                                       (uint32_t) KnownRegisters::BSM_COUNTER_0 +
                                                                                                       counter, 0)
                                                                                               << ", <"
                                                                                               << context.getLabel(
                                                                                                       jumpAddress)
                                                                                               << ">").str()},
    };
}

std::vector<uint32_t> Instruction::Loop::getPossibleBranches() const {
    return std::vector<uint32_t>{_endAddress, jumpAddress};
}

std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>>
Instruction::Loop::execute(AnalysisState &state) const {
    std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>> branches;

    uint32_t counterAddress = (uint32_t) KnownRegisters::BSM_COUNTER_0 + counter;

    if (state.getRegister(counterAddress) == 0) {
        state.current = _endAddress;

        std::unordered_map<uint32_t, uint32_t> branchedState;
        branchedState[counterAddress] = 1;
        branches.emplace_back(jumpAddress, std::move(branchedState));
    } else {
        state.setRegister(counterAddress, state.getRegister(counterAddress) - 1);
        state.current = jumpAddress;

        std::unordered_map<uint32_t, uint32_t> branchedState;
        branchedState[counterAddress] = 0;
        branches.emplace_back(_endAddress, std::move(branchedState));
    }

    state.addKnownJump(getEndAddress(), getAddress(), JumpKind::Continue);
    state.addKnownJump(jumpAddress, getAddress(), JumpKind::Loop);

    return branches;
}

std::vector<uint8_t> Instruction::Loop::toBytes() const {
    std::vector<uint8_t> bytes;

    bytes.emplace_back((uint8_t) getCommand() | counter);

    bytes.emplace_back((jumpAddress >> 16) & 0xFF);
    bytes.emplace_back((jumpAddress >> 8) & 0xFF);
    bytes.emplace_back(jumpAddress & 0xFF);

    return bytes;
}
