/*****************************************************************************
 * Copyright (c) 2020, rrcSmall FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "Set.h"
#include "../Registers.h"
#include <iostream>
#include <iomanip>
#include "../OutputContext.h"

void Instruction::Set::fromBytes(uint32_t offset, const std::vector<uint8_t> &bytes) {
    _address = offset;

    uint8_t command = bytes[offset++];

    address.offset = command & 0b11;

    address.address = bytes[offset++] << 16;
    address.address |= bytes[offset++] << 8;
    address.address |= bytes[offset++];

    value = bytes[offset++] << 24;
    value |= bytes[offset++] << 16;
    value |= bytes[offset++] << 8;
    value |= bytes[offset++];

    mask = bytes[offset++] << 24;
    mask |= bytes[offset++] << 16;
    mask |= bytes[offset++] << 8;
    mask |= bytes[offset++];


    _endAddress = offset;
}

std::string Instruction::Set::toString() const {
    std::stringstream op;
    op << "SET " << getAddressRegisterName(address) << " = ("
       << getAddressRegisterName(address) << " & ~0x" << std::hex << std::setw(8) << std::setfill('0')
       << mask << ") | " << "0x" << std::hex << std::setw(8) << std::setfill('0') << value;
    return op.str();
}

std::vector<Instruction::OutputFormat> Instruction::Set::toOutputFormat(const OutputContext &context) const {
    return std::vector<OutputFormat>{
            OutputFormat{getAddress(), dynamic_cast<std::stringstream &>(std::stringstream("") << "SET" << " "
                                                                                               << context.getAddressRegisterName(
                                                                                                       address)).str()},
            OutputFormat{getAddress() + 4,
                         dynamic_cast<std::stringstream &>(std::stringstream("") << "    VALUE 0x" << std::hex
                                                                                 << std::setw(8) << std::setfill('0')
                                                                                 << value).str()},
            OutputFormat{getAddress() + 8,
                         dynamic_cast<std::stringstream &>(std::stringstream("") << "    MASK 0x" << std::hex
                                                                                 << std::setw(8) << std::setfill('0')
                                                                                 << mask).str()}
    };
}

std::vector<uint32_t> Instruction::Set::getPossibleBranches() const {
    return std::vector<uint32_t>{_endAddress};
}

std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>>
Instruction::Set::execute(AnalysisState &state) const {
    state.setRegister(address, (state.getRegister(address) & ~mask) | value);

    state.current = _endAddress;

    state.addKnownJump(getEndAddress(), getAddress(), JumpKind::Continue);

    return std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>>();
}

std::vector<uint8_t> Instruction::Set::toBytes() const {
    std::vector<uint8_t> bytes;

    bytes.emplace_back((uint8_t) getCommand() | address.offset);

    bytes.emplace_back((address.address >> 16) & 0xFF);
    bytes.emplace_back((address.address >> 8) & 0xFF);
    bytes.emplace_back(address.address & 0xFF);

    bytes.emplace_back((value >> 24) & 0xFF);
    bytes.emplace_back((value >> 16) & 0xFF);
    bytes.emplace_back((value >> 8) & 0xFF);
    bytes.emplace_back(value & 0xFF);

    bytes.emplace_back((mask >> 24) & 0xFF);
    bytes.emplace_back((mask >> 16) & 0xFF);
    bytes.emplace_back((mask >> 8) & 0xFF);
    bytes.emplace_back(mask & 0xFF);

    return bytes;
}
