/*****************************************************************************
 * Copyright (c) 2020, rrcSmall FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "Write.h"
#include "../Registers.h"
#include <iostream>
#include <iomanip>
#include "../OutputContext.h"

void Instruction::Write::fromBytes(uint32_t offset, const std::vector<uint8_t> &bytes) {
    _address = offset;

    uint32_t command = bytes[offset++];

    uint8_t count = (1 + (command >> 2) & 0b1111);
    address.offset = command & 0b11;

    address.address = bytes[offset++] << 16;
    address.address |= bytes[offset++] << 8;
    address.address |= bytes[offset++];

    for (uint32_t i = 0; i < count; ++i) {
        uint32_t entry = bytes[offset++] << 24;
        entry |= bytes[offset++] << 16;
        entry |= bytes[offset++] << 8;
        entry |= bytes[offset++];
        data.emplace_back(entry);
    }


    _endAddress = offset;
}

std::string Instruction::Write::toString() const {
    std::stringstream op;
    op << "WRITE " << getAddressRegisterName(address) << " = 0x";

    for (auto d : data) {
        op << std::hex << std::setw(8) << std::setfill('0') << d;
    }
    return op.str();
}

std::vector<Instruction::OutputFormat> Instruction::Write::toOutputFormat(const OutputContext &context) const {
    uint32_t offset = 0;

    std::vector<OutputFormat> f{
            OutputFormat{getAddress() + offset++ * 4,
                         dynamic_cast<std::stringstream &>(std::stringstream("") << "WRITE" << " "
                                                                                 << context.getAddressRegisterName(address) << ", "
                                                                                 << std::dec
                                                                                 << (uint32_t) data.size()).str()}
    };

    for (auto d : data) {
        f.emplace_back(OutputFormat{getAddress() + offset * 4,
                                    dynamic_cast<std::stringstream &>(std::stringstream("") << "    "
                                                                                            << context.getAddressRegisterName(
                                                                                                    address.address + offset - 1, address.offset)
                                                                                            << " = 0x" << std::hex
                                                                                            << std::setw(8)
                                                                                            << std::setfill('0')
                                                                                            << d).str()});
        offset++;
    }


    return f;
}

std::vector<uint32_t> Instruction::Write::getPossibleBranches() const {
    return std::vector<uint32_t>{_endAddress};
}

std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>>
Instruction::Write::execute(AnalysisState &state) const {
    auto memoryOffset = state.getAddressOffset(address.offset);

    for (uint32_t i = 0; i < data.size(); ++i) {
        state.setRegister(memoryOffset + address.address + i, data[i]);
    }

    state.current = _endAddress;

    state.addKnownJump(getEndAddress(), getAddress(), JumpKind::Continue);

    return std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>>();
}

std::vector<uint8_t> Instruction::Write::toBytes() const {
    std::vector<uint8_t> bytes;

    bytes.emplace_back((uint8_t) getCommand() | ((data.size() - 1) << 2) | address.offset);

    bytes.emplace_back((address.address >> 16) & 0xFF);
    bytes.emplace_back((address.address >> 8) & 0xFF);
    bytes.emplace_back(address.address & 0xFF);

    for (auto d : data) {
        bytes.emplace_back((d >> 24) & 0xFF);
        bytes.emplace_back((d >> 16) & 0xFF);
        bytes.emplace_back((d >> 8) & 0xFF);
        bytes.emplace_back(d & 0xFF);
    }

    return bytes;
}
